// C++ implementation of galois field arithmetic
// January 1999
// Copyright 1999 Phil Karn
// May be used under the terms of the GNU Public License (GPL)

#ifndef _GALOIS_H
#define _GALOIS_H

#define DEBUG 1 // Set to enable extensive run-time error checking

// These two parameters define the field size; make sure they're consistent!
#define MM 8 // base 2 log of field size
#define NN 255 //2**MM - 1

#define MAXCOEFF 256 // The maximum degree of a polynomial

#include <stdio.h>
#include <iostream.h>

#define min(x,y) ((x) < (y) ? (x) : (y))
#define max(x,y) ((x) > (y) ? (x) : (y))

void init_galois(void);

/* Compute x % NN, where NN is 2**MM - 1,
 * without a slow divide
 */
static inline int
modnn(int x)
{
  while (x >= NN) {
    x -= NN;
    x = (x >> MM) + (x & NN);
  }
  return x;
}

/* Field generator polynomial */
extern int Pp[];

/* index->polynomial form conversion table */
extern int Alpha_to[];

/* Polynomial->index form conversion table */
extern int Index_of[];

/* No legal value in index form represents zero, so
 * we need a special value for this purpose
 */
extern const int A0;

enum form {POLY,INDEX};

class galois
{
private:
  int poly_value;

public:
  // create galois element
  galois(void);
  // create galois element with specific value assumed in poly form
  galois(const int x);
  // create galois element with specified value in index form
  friend galois galois_index(const int x);
  // copy
  galois(const galois& arg);
  // return galois element in polynomial form
  int poly(void) const;
  // return galois element in index form
  int index(void) const ;
  // equals (assignment)
  galois operator =(const galois arg);
  galois operator =(const int x);
  // binary operators that modify their first arg
  galois& operator +=(const galois& b);
  galois& operator +=(const int b);
  galois& operator -=(const galois& b);
  galois& operator *=(const galois& b);
  galois& operator ^=(const int b);
  galois& operator /=(const galois &b);
  galois inv(void);
};  
// create galois element with undefined value
#ifdef DEBUG
inline galois::galois(void){ poly_value = -1;}
#else
inline galois::galois(void){ }
#endif
// create galois element with specific value in poly form
inline galois::galois(const int x){
  poly_value = x & NN;
}
// create galois element with specified value in index form
inline galois galois_index(const int x){
  galois result;
  result.poly_value = Alpha_to[x];
  return result;
}
// copy
inline galois::galois(const galois& arg) {
#ifdef DEBUG
  if(arg.poly_value == -1){
    printf("undefined value\n");
    abort();
  }
#endif
  poly_value = arg.poly_value;
}
// return galois element in polynomial form
inline int galois::poly(void) const {
  return poly_value;
}
// return galois element in index form
inline int galois::index(void) const {
#ifdef DEBUG
  if(poly_value == -1){
    printf("undefined value\n");
    abort();
  }
#endif
  return Index_of[poly_value];
}
// equals (assignment)
inline galois galois::operator =(const galois arg) {
#ifdef DEBUG
  if(arg.poly_value == -1){
    printf("undefined value\n");
    abort();
  }
#endif
  poly_value = arg.poly_value;
}
inline galois galois::operator =(const int x) {
  poly_value = x & NN;
}

// comparisons
inline int operator ==(const galois& arg1,const galois& arg2) {
  return (arg1.poly() == arg2.poly());
}
inline int operator ==(const galois& arg1,const int arg2) {
  return (arg1.poly() == arg2);
}
inline int operator ==(const int arg1,const galois& arg2) {
  return (arg1 == arg2.poly());
}
inline int operator !=(const galois& arg1,const galois& arg2) {
  return (arg1.poly() != arg2.poly());
}
inline int operator !=(const galois& arg1,const int arg2) {
  return (arg1.poly() != arg2);
}
inline int operator !=(const int arg1,galois& arg2) {
  return (arg1 != arg2.poly());
}

// Galois field addition is bitwise XOR
inline galois operator +(const galois& a,const galois& b) {
  return galois(a.poly() ^ b.poly());
}
inline galois operator +(const galois& a,const int b) {
  return galois(a.poly() ^ b);
}
inline galois operator +(const int a,const galois& b) {
  return galois(a ^ b.poly());
}
inline galois& galois::operator +=(const galois& b) {
  *this = *this + b;
  return *this;
}
inline galois& galois::operator +=(const int b) {
  *this = *this + b;
  return *this;
}

// Galois field subtraction is the same as addition
inline galois operator -(const galois& a,const galois& b) {
  return galois(a.poly() ^ b.poly());
}
inline galois operator -(const galois& a,const int b) {
  return galois(a.poly() ^ b);
}
inline galois operator -(const int a,const galois& b) {
  return galois(a ^ b.poly());
}
inline galois& galois::operator -=(const galois& b) {
  *this = *this - b;
  return *this;
}

// Additive inverse element is the same as the element
inline galois operator -(const galois& a) {
  return a;
}

// Multiplicative inverse
inline galois galois::inv(void) {
#ifdef DEBUG
  if(poly() == 0){
    printf("Inverse of zero element\n");
    abort();
  }
#endif
  return galois_index(modnn(NN - index()));
}

// Galois field multiplication
inline galois operator *(const galois& a,const galois& b) {
  if(a == 0 || b == 0)
    return galois(0);
  return galois_index(modnn(a.index() + b.index()));
}
inline galois& galois::operator *=(const galois& b){
  *this = *this * b;
  return *this;
}

// Galois field exponentiation
inline galois operator ^(const galois& a, int b) {
  if(a == 0)
    return 0;
  while(b < 0)
    b += NN;// b could be negative
  if(b == 0)
    return 1;
  return galois_index(modnn(a.index() * b));
}
inline galois& galois::operator ^=(const int b){
  *this = *this ^ b;
}

// Galois field division
inline galois operator /(const galois& a,const galois& b) {
  if(a == 0)
    return a;
#ifdef DEBUG
  if(b == 0){
    printf("Error: galois division by zero\n");
    abort();
  }
#endif
  return galois_index(modnn(a.index() - b.index() + NN));
}
inline galois& galois::operator /=(const galois &b) {
  *this = *this / b;
  return *this;
}  


// Primitive root alpha, can be used in expressions like
// ALPHA^5 = alpha to the fifth power
extern const galois ALPHA;

// Polynomials over galois fields
class polynomial {
private:
  int deg;
  galois coeffs[MAXCOEFF];

public:
  polynomial(void);
  // Construct polynomial of specified degree with zeroed coefficients
  polynomial(const int n);
  // Construct 0th degree polynomial from galois constant
  polynomial(const galois& x);
  // Construct polynomial of specified degree
  polynomial(const int n,const galois c[]);
  // Construct polynomial of specified degree with specified coefficients
  polynomial(const int n,const enum form& form,...);
  // Clear polynomial, setting specified degree
  void clear(int d);
  // Copy polynomial
  polynomial(const polynomial& arg);
  // assignment operator
  polynomial& operator =(const polynomial& arg);
  polynomial& operator =(const int x);
  // Extract or set coefficient
  galois& operator [](const int n);
  galois operator [](const int n) const;
  // set degree without affecting coeffs
  inline void set_degree(const int& x);

  // no op
  ~polynomial(void);
  // Degree of polynomial
  int degree(void) const;
  // Polynomial evaluation
  galois operator ()(const galois& arg) const;
  galois operator ()(const int x) const;
  // Polynomial addition
  polynomial& operator +=(const polynomial& arg);
  polynomial& operator +=(const galois& x);
  polynomial& operator +=(const int x);
  // Polynomial subtraction is the same as addition
  polynomial& operator -=(const polynomial& arg);
  polynomial& operator -=(const galois& arg);
  polynomial& operator -=(const int arg);
  // Polynomial multiplication
  polynomial& operator *=(const polynomial& multiplier);
  polynomial& operator *=(const galois& x);
  polynomial& operator *=(const int x);
  // Polynomial division
  polynomial& operator /=(const polynomial& divisor);
  polynomial& operator /=(const galois& divisor);
  polynomial& operator /=(const int divisor);
  // Polynomial multiplication by X^n
  polynomial operator <<=(const int n);
  // Polynomial exponentation
  polynomial operator ^=(const int n);
  // Polynomial division, returning remainder
  polynomial& operator %=(const polynomial& divisor);
};

// Constant X that may be used to construct polynomials by hand, e.g.,
// (X<<2) + 2*(X<<1) + 3*(X) + 1
extern const polynomial X;

inline polynomial::polynomial(void){
  deg = -1;
}
// Construct polynomial of specified degree
inline polynomial::polynomial(const int n){
#if DEBUG
  if(n >= MAXCOEFF){
    printf("error: degree %d > MAXCOEFF=%d\n",n,MAXCOEFF);
    abort();
  }
#endif
  deg = n;
}
// Construct polynomial of specified degree with specified coefficients
inline polynomial::polynomial(const int n,const galois c[]){
  int i;
#if DEBUG
  if(n >= MAXCOEFF){
    printf("error: degree %d > MAXCOEFF=%d\n",n,MAXCOEFF);
    abort();
  }
#endif
  deg = n;
  for(i=deg;i>=0;i--)
    coeffs[i] = c[i];
}

// Construct polynomial of specified degree with specified coefficients
inline polynomial::polynomial(const int n,const enum form& form,...){
  int i;
  va_list ap;

#if DEBUG
  if(n >= MAXCOEFF){
    printf("error: degree %d > MAXCOEFF=%d\n",n,MAXCOEFF);
    abort();
  }
#endif

  va_start(ap,form);
  
  deg = n;
  switch(form){
  case INDEX:
    for(i=0;i<=deg;i++){
      coeffs[i] = galois_index(va_arg(ap,int));
    }
    break;
  case POLY:
    for(i=0;i<=deg;i++){
      coeffs[i] = galois(va_arg(ap,int));
    }
    break;
  }
  va_end(ap);
}

inline polynomial::polynomial(const galois& x){
  deg = 0;
  coeffs[0] = x;
}

// Clear polynomial, setting specified degree
inline void polynomial::clear(int d){
  int i;
  
#if DEBUG
  if(d >= MAXCOEFF){
    printf("error: degree %d > MAXCOEFF=%d\n",d,MAXCOEFF);
    abort();
  }
#endif

  deg = d;
  for(i=deg;i>=0;i--){
    this->coeffs[i] = 0;
  }
}
// Copy polynomial
inline polynomial::polynomial(const polynomial& arg) {
  int i;
  deg = arg.deg;
  for(i=deg;i>=0;i--)
    coeffs[i] = arg.coeffs[i];
}
// assignment operator
inline polynomial& polynomial::operator =(const polynomial& arg){
  int i;
  deg = arg.deg;
  for(i=deg;i>=0;i--)
    coeffs[i] = arg.coeffs[i];
  return *this;
}
inline polynomial& polynomial::operator =(const int x){
  deg = 0;
  coeffs[0] = x;
}
// set degree without affecting coeffs
inline void polynomial::set_degree(const int& x) {
  deg = x;
}

// return coefficient reference for use as lvalue
inline galois& polynomial::operator [](const int n) {
#ifdef DEBUG
  if(n < 0 || n > deg){
    printf("out of range coefficient!\n");
    abort();
  }
#endif
  return coeffs[n];
}
// return coefficient for use as rvalue
inline galois polynomial::operator [](const int n)const {
#ifdef DEBUG
  if(n < 0 || n > deg){
    printf("out of range coefficient!\n");
    abort();
  }
#endif
  return coeffs[n];
}
// Degree of polynomial
inline int polynomial::degree(void) const {
  return deg;
}

// no op
inline polynomial::~polynomial(void){}

// Polynomial evaluation
inline galois polynomial::operator ()(const galois& arg) const {
  int i;
  galois result;
  
  if(deg < 0)
    return 0;
  result = (*this)[deg];
  for(i=deg-1;i>=0;i--){
    result = (*this)[i] + (result * arg);
  }
  return result;
}
inline galois polynomial::operator ()(const int x) const {
  return (*this)(galois(x));
}

// Comparision
int operator ==(const polynomial& arg1,const polynomial& arg2);
// Polynomial addition
polynomial operator +(const polynomial& arg1,const polynomial& arg2);
polynomial operator +(const polynomial& arg1,const galois& arg2);
// Polynomial multiplication
polynomial operator *(const polynomial& arg1,const polynomial& arg2);
polynomial operator *(const polynomial& arg1,const galois& arg2);
// Polynomial division, returning quotient
polynomial operator /(const polynomial& dividend,const polynomial& divisor);
polynomial operator /(const polynomial& dividend,const galois& divisor);
// Multiply or divide polynomial by X^n
polynomial operator <<(const polynomial& arg1,const int arg2);
polynomial operator >>(const polynomial& arg1,const int arg2);
// Polynomial division, returning remainder
polynomial operator %(const polynomial& dividend,const polynomial& divisor);

inline int operator ==(const polynomial& arg1,const int arg2) {
  if((arg1.degree() == 0) && (arg1[0] == arg2))
    return 1;
  return 0;
}
inline int operator ==(const int arg1,const polynomial& arg2) {
  if((arg2.degree() == 0) && (arg1 == arg2[0]))
    return 1;
  return 0;
}
inline int operator !=(const polynomial& arg1,const polynomial& arg2) {
  return !(arg1 == arg2);
}
inline int operator !=(const polynomial& arg1,const int arg2) {
  if((arg1.degree() != 0) || (arg1[0] != arg2))
    return 1;
  return 0;
}
inline int operator !=(const int arg1,const polynomial& arg2) {
  if((arg2.degree() != 0) || (arg2[0] != arg1))
    return 1;
  return 0;
}

inline polynomial operator +(const polynomial& arg1,const int arg2) {
  return arg1 + galois(arg2);
}
inline polynomial operator +(const galois& arg1,const polynomial& arg2){
  return arg2 + arg1;
}
inline polynomial operator +(const int arg1,const polynomial& arg2) {
  return arg2 + arg1;
}

inline polynomial& polynomial::operator +=(const polynomial& arg) {
  *this = *this + arg;
  return *this;
}
inline polynomial& polynomial::operator +=(const galois& arg) {
  *this = *this + arg;
  return *this;
}
inline polynomial& polynomial::operator +=(const int arg) {
  *this = *this + arg;
  return *this;
}

// Polynomial subtraction is the same as addition
inline polynomial operator -(const polynomial& arg1,const polynomial& arg2){
  return arg1 + arg2;
}
inline polynomial operator -(const polynomial& arg1,const galois& arg2){
  return arg1 + arg2;
}
inline polynomial operator -(const polynomial& arg1,const int arg2){
  return arg1 + arg2;
}
inline polynomial operator -(const galois& arg1,const polynomial& arg2){
  return arg1 + arg2;
}
inline polynomial operator -(const int arg1,const polynomial& arg2){
  return arg1 + arg2;
}
inline polynomial& polynomial::operator -=(const polynomial& arg) {
  *this = *this + arg;
  return *this;
}
inline polynomial& polynomial::operator -=(const galois& x) {
  *this = *this + x;
  return *this;
}
inline polynomial& polynomial::operator -=(const int x) {
  *this = *this + x;
  return *this;
}


inline polynomial operator *(const polynomial& arg1,const int arg2){
  return arg1 * galois(arg2);
}
inline polynomial operator *(const galois& arg1,const polynomial& arg2) {
  return arg2 * arg1;
}
inline polynomial operator *(const int arg1,const polynomial& arg2) {
  return arg2 * galois(arg1);
}
inline polynomial& polynomial::operator *=(const polynomial& multiplier){
  *this = *this * multiplier;
  return *this;
}
inline polynomial& polynomial::operator *=(const galois& multiplier){
  *this = *this * multiplier;
  return *this;
}
inline polynomial& polynomial::operator *=(const int multiplier){
  *this = *this * multiplier;
  return *this;
}

inline polynomial operator /(const polynomial& dividend,const int divisor){
  return dividend / galois(divisor);
}
inline polynomial& polynomial::operator /=(const polynomial& divisor){
  *this = *this / divisor;
  return *this;
}
inline polynomial& polynomial::operator /=(const galois& divisor){
  *this = *this / divisor;
  return *this;
}
inline polynomial& polynomial::operator /=(const int divisor){
  *this = *this / divisor;
  return *this;
}

inline polynomial polynomial::operator <<=(const int n) {
  *this = *this << n;
  return *this;
}

polynomial operator ^(const polynomial& a, const int n);
inline polynomial polynomial::operator ^=(const int n) {
  *this = *this ^ n;
  return *this;
}

inline polynomial& polynomial::operator %=(const polynomial& divisor){
  *this = *this % divisor;
  return *this;
}

// Constant X that may be used to construct polynomials by hand, e.g.,
// (X<<2) + 2*(X<<1) + 3*(X) + 1
extern const polynomial X;

// display polynomial in polynomial form
void dump_poly(const polynomial& p,const char *label);
// display in index form
void dump_index(const polynomial& p,const char *label);

#endif
